package sayu;
import java.util.*;

/**
   @author Jesse Davis

   This class contains some Utility methods need for either learning a
   Tree Augmented Bayes from scratch or incrementally.

   (c) 2010
 */
public class GraphUtils{

    /**
       Kruskal's algorithm for finding a maximum weigth spanning tree. 

       @param tree are the possible edges to be included in the tree
       @param pickedEdges are the edges picked for the spanning tree
       @param graph is an adjaceny matrix representation of the picked tree
       @param outDegrees contains the outDegree of each node in the graph
     */
    public static void kruskals(Edge[] tree, int[][] graph,  int[] outDegrees, Edge[] pickedEdges){
        Arrays.sort(tree);
   
        int v1, v2;
        //go through and try adding edges
        double score = 0;
        int curr = pickedEdges.length-1;
        for(int ix = (tree.length -1); ix >=0 && curr >= 0; ix--){
            //get information to add edge
            v1 = tree[ix].getStart();
            v2 = tree[ix].getEnd();
 
            //add the edge
            graph[v1][v2] = Constants.IN_EDGE;
            graph[v2][v1] = Constants.IN_EDGE;
            outDegrees[v1]++;
            outDegrees[v2]++;
            //if add is invalid remove the edge
            if (!acyclic(graph, outDegrees)){
                graph[v1][v2] = Constants.NO_EDGE;
                graph[v2][v1] = Constants.NO_EDGE;
                outDegrees[v1]--;
                outDegrees[v2]--;
            }
            else{
                pickedEdges[curr] = tree[ix];
		
		//System.out.print("inc TAN Picked Edge " + tree[ix].toString());
		
                curr--;
                score+=tree[ix].getWeight();
            }
        }//end of ix loop

      
    }//end of method kruskal


    /**
       Version of Kruskal's algorithm for finding a maximum weight
       spanning tree that works with incremental TAN.

       @param bestEdges are the edges currently in the model. Should be sorted.
       @param evalEdges are new edges that could be added to the tree
       @param currEdges are the edges picked for the spanning tree
       @param graph is an adjaceny matrix representation of the picked tree
       @param outDegrees contains the outDegree of each node in the graph
     */
    public static void kruskals(Edge[] bestEdges, Edge[] evalEdges, Edge[] currEdges, int[][] graph, int[] outDegrees){
        Arrays.sort(evalEdges);
	//FLAG 
        int v1, v2;
        int position = currEdges.length-1;
        int currEdgePos = bestEdges.length-1; 
        int evalEdgePos = evalEdges.length-1;
 
        double ipEval = 0, ipCurr = 0;

        int edgeCount = 0;
	Edge pickedEdge;
	
        //construct the graph
        while (position >= 0){

            if (evalEdgePos >= 0  && currEdgePos >= 0){
                ipEval = evalEdges[evalEdgePos].getWeight();
                ipCurr = bestEdges[currEdgePos].getWeight();
                //take edge involving new rule
                if (ipEval > ipCurr){
                    v1 = evalEdges[evalEdgePos].getStart();
                    v2 = evalEdges[evalEdgePos].getEnd();
		    pickedEdge = evalEdges[evalEdgePos];
                    evalEdgePos--;
                }
                //take edge involving only old rules
                else{
                    v1 = bestEdges[currEdgePos].getStart();
                    v2 = bestEdges[currEdgePos].getEnd();
		    pickedEdge = bestEdges[currEdgePos];
                    currEdgePos--;
                }
            }
            //used all edges involving only old rules
            else if (evalEdgePos >= 0 && currEdgePos < 0){
                v1 = evalEdges[evalEdgePos].getStart();
                v2 = evalEdges[evalEdgePos].getEnd();
		pickedEdge = evalEdges[evalEdgePos];
                evalEdgePos--;
            }
            //used all edges involving new rule
            else{
                v1 = bestEdges[currEdgePos].getStart();
                v2 = bestEdges[currEdgePos].getEnd();
		pickedEdge = bestEdges[currEdgePos];
                currEdgePos--;
            }
            //Speculatively add the edge to the graph
            graph[v1][v2] = Constants.IN_EDGE;
            graph[v2][v1] = Constants.IN_EDGE;
            outDegrees[v1]++;
            outDegrees[v2]++;

            //if edge results in a undirected cycle, remove it
            if (!acyclic(graph, outDegrees)){
                graph[v1][v2] = Constants.NO_EDGE;
                graph[v2][v1] = Constants.NO_EDGE;
                outDegrees[v1]--;
                outDegrees[v2]--;
            }
            else{
				//System.out.println("inc TAN Picked Edge " + pickedEdge.toString());
                currEdges[position] = pickedEdge;
                position--;
            }
        }//end of loop
    }//end method kruskal

    /**
       This algorithm is for a directed graph. It checks to see if the graph
       has any cycles in it.

       I got this algorithm from: 
       http://www.cs.hmc.edu/~keller/courses/cs60/s98/examples/acyclic/

       @param graph is an adjaceny matrix representation of the graph
       @param outDegrees contains the outDegree of each node in the graph
       @return true if graph is acyclic
     */
    public static boolean acyclic(int[][] graph, int[] outDegrees){

      
	//FLAG
        int vertex = graph.length;
        boolean failed = false;
	//FLAG
        boolean[] valid = new boolean[graph.length];
	//FLAG
        int[] tempOutDegrees = (int[]) outDegrees.clone();
	//FLAG
        for (int i = 0; i < valid.length; i++){
            valid[i] = true;
        }
        while ((vertex > 0) && (!failed)){
            int pos = 0;
            boolean hasLeaf = false;
	    //Flag
            for(; pos < graph.length && !hasLeaf; pos++){

		if ((tempOutDegrees[pos] <= 1 ) && (valid[pos])){
		    hasLeaf = true;
		    vertex--;
		    valid[pos] = false;
		}
            }

            if (!hasLeaf){               
                failed = true;
                continue;
            }

            /*
	      remove the vertex at pos from the graph and all the edges
	      entering the vertex
             */
            pos--;
            for(int ix = 0; ix < graph.length && pos < graph.length; ix++){
                if (graph[pos][ix] == Constants.IN_EDGE){
                    tempOutDegrees[ix]--;
                }
            }//end for

        }//end while

        return !failed;
    }//end method

    /**
       Turns an undirected spanning tree into a directed graph.
       @param graph is a undirected graph that representes a spanning tree. This will be converted into a directed graph.
       @param root the current node being evaluated.
       @param outDegrees the outDegree of each vertex in the graph
       @param parent the vertex that is the parent of root
       @param pars the non-class parent of each vertex in the graph.  Note that in a TAN structure, each node can have at most one additional parent other than the class variable. 
     */
    public static void makeDirected(int[][] network, int root, int[] outDegrees,
				    int parent, int[] inDegrees, int[] pars){

        if (outDegrees[root] > 0){
            for(int ix = 0; ix < network.length; ix++){
                if (network[root][ix] != Constants.NO_EDGE &&
                        ix != parent){
                    outDegrees[ix]--;
                    network[root][ix] = Constants.OUT_EDGE;
                    inDegrees[ix]++;
                    pars[ix] = root;
                    makeDirected(network, ix, outDegrees, root, inDegrees, pars);
                }
            }
        }//end of 
    }//end method makeDirected
}//end of class GraphUtils