package sayu;

/**
   @author Jesse Davis
   @author Vitor Santos Costa

   This is the interface that we have defined for use with the SAYU
   and VISTA algorithms for constructing a propositional classifier
   whose features are first-order Horn-clauses.

   (c) 2010
 */
public interface ILPSearch{
  
    /**
       This method introduces a new attribute into the classifier.  It
       relearns a model incorporating this feature, then evaluates the
       new classifier on the tune set according to the score function
       the user selected.

       @param trainPosVector value of new attribute for the pos train examples
       @param trainNegVector value of new attribute for the neg train examples
       @param tunePosVector value of new attribute for the pos tune examples
       @param tuneNegVector value of new attribute for the neg tune examples
       @param testPosVector value of new attribute for the pos test examples
       @param testNegVector value of new attribute for the neg test examples
       @param scoreFunction metric for scoring the new network
       @param numValues the number of values the new attribute can take
       @return the score of the model incorporating the new feature
     */
    public double addAttribute(byte[] trainPosVector, byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector, byte[] testNegVector, int scoreFunction, int numValues);


    /**
      This method introduces a new attribute into the network.  It
      relearns a BN incorporating this feature, then evaluates the
      new network on the tune set according to the score function the
      user selected.

      @param trainPosVector value of new attribute for the pos train
examples
      @param trainNegVector value of new attribute for the neg train
examples
      @param tunePosVector value of new attribute for the pos tune examples
      @param tuneNegVector value of new attribute for the neg tune examples
      @param testPosVector value of new attribute for the pos test examples
      @param testNegVector value of new attribute for the neg test examples
      @param scoreFunction metric for scoring the new network
      @param numValues the number of values the new attribute can take
      @return the score of the model incorporating the new feature
    */
    public double[] addAttribute2(byte[] trainPosVector, byte[]
				  trainNegVector, byte[] tunePosVector, byte[] tuneNegVector, byte[]
				  testPosVector, byte[] testNegVector, int scoreFunction, int numValues);

    /**
       This method removes an attribute from the classifier.  It
       relearns a TAN model without this feature, then evaluates the
       new model on the tune set according to the score function the
       user selected.

       @param removePredNum the predicate number to removed from the model
       @param scoreFunction metric for scoring the new network
       @return the score of the model incorporating the new feature
     */
    public double removeAttribute(int removePredNum, int scoreFunction);

    /**
       This method replaces an attribute already in the model. It
       relearns a classifier incorporating this feature in place of a
       previous feature, then evaluates the new model on the tune set
       according to the score function the user selected.

       @param trainPosVector value of new attribute for the pos train examples
       @param trainNegVector value of new attribute for the neg train examples
       @param tunePosVector value of new attribute for the pos tune examples
       @param tuneNegVector value of new attribute for the neg tune examples
       @param testPosVector value of new attribute for the pos test examples
       @param testNegVector value of new attribute for the neg test examples
       @param replacePredNum the predicate number to replaced in the model
       @param scoreFunction metric for scoring the new network
       @param numValues the number of values the new attribute can take
       @return the score of the model incorporating the new feature
     */
    public double replaceAttribute(byte[] trainPosVector, byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector, byte[] testNegVector, int replacePredNum, int scoreFunction, int numValues);

    /**
       This method permanently adds the most recently introduced
       feature into feature set.  Updates the cpts and network
       topology to reflect the change.
     */
    public void commit();

    /**
       Scores the current model on either train, tune or test
       set. Currently, the following score functions are
       available: CLL, AUC-PR, AUC-ROC and Accuracy.

       @param dataSet whether to score the train, tune or test set
       @param scoreFunction the metric to use for scoring the classifier
       @return score of the model
     */  
    public double getScore(int dataSet, int scoreFunction);

    /**
       Calculates the probability that each example in either the
       train, tune or test set belongs to the positive class. It uses
       the current model structure.

       @param dataSet whether to return the probabilities for the train, tune or test set.
       @return Prob(Pos Class) for each example in the desired dataSet
     */
    public double[] getProbs(int dataSet);

    /**
       @return number of non-class attributes in the model
     */
    public int getNumRules();

    /**
       Called when constructing a new predicate head. 

       @param type an unique id to present the type and order of
       variables in the predicate head being learned.
     */  
    public void newHead(int type);

    /**
       Sets the recall threshold to start integrating from when
       calculating the area under the precision-recall curve.

       @param minRecall is the recall level at which to begin calculating the AUC-PR
     */
    public void setAUCPRMinRecall(double minRecall);

    /**
       set the threshold for labeling an example as positive. 

       @param threshold the probability above each an example is classified as positive.
     */
    public void setACCThreshold(double threshold);

    /**
       Prints out statistics from the most recent SAYU run
     */
    public void terminate();

}//end interface ILPSearch