package sayu;

/**
 * @author Jesse Davis
 * 
 *         This class is for incrementally building a Naive Bayes model. It
 *         stores the constructed train, tune and test sets. It also stores the
 *         log probability for each node in the model.
 * 
 *         (c) 2010
 */
public class IncNB extends BNILPSearch {

	// for storing log[prob(val | class)] for each feature
	// feature index by feature value by class value
	private double[][][] logCondProbs;

	// for computing the conditional distribution
	private double[] totals = new double[2];

	// log [Prob(val | neg) - Prob(val | pos) for each attribute, value
	private double[][] ratio;

	/*
	 * The log odds (neg/pos) of each train, tune and test example according to
	 * the current model.
	 */
	private double[] trainLikelihood;
	private double[] tuneLikelihood;
	private double[] testLikelihood;

	/*
	 * The log odds (neg/pos) of each train, tune and test example according to
	 * the current model.
	 */
	private double[] trainEval;
	private double[] tuneEval;
	private double[] testEval;

	/**
	 * Constructs a Naive Bayes model with only the class variable.
	 * 
	 * @param trainPos
	 *            the number of positive examples in the train set.
	 * @param trainNeg
	 *            the number of negative examples in the train set.
	 * @param tunePos
	 *            the number of positive examples in the tune set.
	 * @param tuneNeg
	 *            the number of negative examples in the tune set.
	 * @param testPos
	 *            the number of positive examples in the test set.
	 * @param testNeg
	 *            the number of negative examples in the test set.
	 */
	public IncNB(int trainPos, int trainNeg, int tunePos, int tuneNeg,
			int testPos, int testNeg) {
		init(trainPos, trainNeg, tunePos, tuneNeg, testPos, testNeg);
	}// end IncNB(int, int, int, int, int, int)

	/**
	 * Constructs a Naive Bayes model based on an initial feature set
	 * 
	 * @param trainPos
	 *            the number of positive examples in the train set.
	 * @param trainNeg
	 *            the number of negative examples in the train set.
	 * @param tunePos
	 *            the number of positive examples in the tune set.
	 * @param tuneNeg
	 *            the number of negative examples in the tune set.
	 * @param testPos
	 *            the number of positive examples in the test set.
	 * @param testNeg
	 *            the number of negative examples in the test set.
	 * @param initTrain
	 *            a initial set of train examples
	 * @param initTune
	 *            a initial set of tune examples
	 * @param initTest
	 *            a initial set of test examples
	 * @param initAttribute
	 *            a initial feature set
	 * @param initOutputPos
	 *            is the index of the output position in the initial feature set
	 */
	public IncNB(int trainPos, int trainNeg, int tunePos, int tuneNeg,
			int testPos, int testNeg, byte[][] initTrain, byte[][] initTune,
			byte[][] initTest, Attribute[] initAttributes, int initOutputPos) {
		while (initSize < (initTrain[0].length - 1)) {
			initSize = initSize * 2;
		}
		init(trainPos, trainNeg, tunePos, tuneNeg, testPos, testNeg);

		Utils.copyByteArray(initTrain, train, initOutputPos);
		Utils.copyByteArray(initTune, tune, initOutputPos);
		Utils.copyByteArray(initTest, test, initOutputPos);
		int idx = 0;
		for (int ix = 0; ix < initAttributes.length; ix++) {
			if (ix != initOutputPos) {
				attributes[idx] = initAttributes[ix];
				numValuesFeat[idx] = attributes[idx].getNumValues();
				idx++;
			}
		}
		numRules = initTrain[0].length;
		numPriorFeats = numRules - 1;
		for (int ix = 0; ix < numRules; ix++) {
			logCondProbs[ix] = Utils.condProb(train, ix, outputPosition,
					numValuesFeat[ix], 2, initialStrength, true);
			ratio[ix] = new double[numValuesFeat[ix]];
			for (int jx = 0; jx < ratio[ix].length; jx++) {
				ratio[ix][jx] = logCondProbs[ix][jx][Constants.NEG]
						- logCondProbs[ix][jx][Constants.POS];
			}
		}

		for (int ix = 0; ix < train.length; ix++) {
			for (int jx = 0; jx < numRules; jx++) {
				trainLikelihood[ix] += ratio[jx][train[ix][jx]];
			}
		}

		for (int ix = 0; ix < tune.length; ix++) {
			for (int jx = 0; jx < numRules; jx++) {
				tuneLikelihood[ix] += ratio[jx][tune[ix][jx]];
			}
		}

		for (int ix = 0; ix < test.length; ix++) {
			for (int jx = 0; jx < numRules; jx++) {
				testLikelihood[ix] += ratio[jx][test[ix][jx]];
			}
		}
	}// end IncNB(int, int, int, int, int, int, byte[][], byte[][], byte[][],
		// Attribute[], int)

	/**
	 * Initializes the class variables and sets up the infrastructure for
	 * incrementally building a Naive Bayes model.
	 * 
	 * @param trainPos
	 *            the number of positive examples in the train set.
	 * @param trainNeg
	 *            the number of negative examples in the train set.
	 * @param tunePos
	 *            the number of positive examples in the tune set.
	 * @param tuneNeg
	 *            the number of negative examples in the tune set.
	 * @param testPos
	 *            the number of positive examples in the test set.
	 * @param testNeg
	 *            the number of negative examples in the test set.
	 */
	protected void init(int trainPos, int trainNeg, int tunePos, int tuneNeg,
			int testPos, int testNeg) {

		super.init(trainPos, trainNeg, tunePos, tuneNeg, testPos, testNeg);
		logCondProbs = new double[initSize][][];
		ratio = new double[initSize][];
		trainLikelihood = new double[train.length];
		tuneLikelihood = new double[tune.length];
		testLikelihood = new double[test.length];
		trainEval = new double[train.length];
		tuneEval = new double[tune.length];
		testEval = new double[test.length];

		double prior = cat0Prior - cat1Prior;
		for (int ix = 0; ix < train.length; ix++) {
			trainLikelihood[ix] = prior;
			trainEval[ix] = prior;
		}
		for (int ix = 0; ix < tune.length; ix++) {
			tuneLikelihood[ix] = prior;
			tuneEval[ix] = prior;
		}
		for (int ix = 0; ix < test.length; ix++) {
			testLikelihood[ix] = prior;
			testEval[ix] = prior;
		}
	}// end init(int, int, int, int, int, int)

	/**
	 * This method introduces a new attribute into the network. It learns a NB
	 * model incorporating this feature, then evaluates the new network on the
	 * tune set according to the score function the user selected.
	 * 
	 * @param trainPosVector
	 *            value of new attribute for the pos train examples
	 * @param trainNegVector
	 *            value of new attribute for the neg train examples
	 * @param tunePosVector
	 *            value of new attribute for the pos tune examples
	 * @param tuneNegVector
	 *            value of new attribute for the neg tune examples
	 * @param testPosVector
	 *            value of new attribute for the pos test examples
	 * @param testNegVector
	 *            value of new attribute for the neg test examples
	 * @param scoreFunction
	 *            metric for scoring the new network
	 * @param numValues
	 *            the number of values the new attribute can take
	 * @return the score of the model incorporating the new feature
	 */
	public double addAttribute(byte[] trainPosVector, byte[] trainNegVector,
			byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector,
			byte[] testNegVector, int scoreFunction, int numValues) {
		moveType = Constants.ADD;
		addEvaluated++;
		numValuesFeat[numRules] = numValues;
		// get the conditional counts
		getCondCount(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector);

		ratio[numRules] = new double[numValues];
		for (int ix = 0; ix < numValues; ix++) {
			ratio[numRules][ix] = logCondProbs[numRules][ix][Constants.NEG]
					- logCondProbs[numRules][ix][Constants.POS];
		}
		for (int ix = 0; ix < train.length; ix++) {
			trainEval[ix] = trainLikelihood[ix]
					+ ratio[numRules][train[ix][numRules]];
		}
		for (int ix = 0; ix < tune.length; ix++) {
			tuneEval[ix] = tuneLikelihood[ix]
					+ ratio[numRules][tune[ix][numRules]];
		}
		for (int ix = 0; ix < test.length; ix++) {
			testEval[ix] = testLikelihood[ix]
					+ ratio[numRules][test[ix][numRules]];
		}
		attributes[numRules] = new VariableConsecInt("R" + numRules, numValues);
		return getScore(Constants.TUNE_SET, scoreFunction, tuneEval);
	}// end addAttribute

	/**
	 * This method introduces a new attribute into the network. It relearns a BN
	 * incorporating this feature, then evaluates the new network on the tune
	 * set according to the score function the user selected.
	 * 
	 * @param trainPosVector
	 *            value of new attribute for the pos train examples
	 * @param trainNegVector
	 *            value of new attribute for the neg train examples
	 * @param tunePosVector
	 *            value of new attribute for the pos tune examples
	 * @param tuneNegVector
	 *            value of new attribute for the neg tune examples
	 * @param testPosVector
	 *            value of new attribute for the pos test examples
	 * @param testNegVector
	 *            value of new attribute for the neg test examples
	 * @param scoreFunction
	 *            metric for scoring the new network
	 * @param numValues
	 *            the number of values the new attribute can take
	 * @return the score of the model incorporating the new feature
	 */
	public double[] addAttribute2(byte[] trainPosVector, byte[] trainNegVector,
			byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector,
			byte[] testNegVector, int scoreFunction, int numValues) {
		double[] out = new double[2];
		out[1] = addAttribute(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector, scoreFunction,
				numValues);
		out[0] = getScore(Constants.TRAIN_SET, scoreFunction, trainEval);
		return out;
	}

	/**
	 * This method removes an attribute from the network. It relearns a NB
	 * network without this feature, then evaluates the new network on the tune
	 * set according to the score function the user selected.
	 * 
	 * @param removePredNum
	 *            the predicate number to removed from the model
	 * @param scoreFunction
	 *            metric for scoring the new network
	 * @return the score of the model incorporating the new feature
	 */
	public double removeAttribute(int removePredNum, int scoreFunction) {
		modifiedV1 = predNameMap.get(removePredNum);
		int removePos = modifiedV1;
		modPred1 = removePredNum;
		moveType = Constants.REMOVE;
		removeEvaluated++;
		for (int ix = 0; ix < train.length; ix++) {
			trainEval[ix] = trainLikelihood[ix]
					- ratio[removePos][train[ix][removePos]];
		}
		for (int ix = 0; ix < tune.length; ix++) {
			tuneEval[ix] = tuneLikelihood[ix]
					- ratio[removePos][tune[ix][removePos]];
		}
		for (int ix = 0; ix < test.length; ix++) {
			testEval[ix] = testLikelihood[ix]
					- ratio[removePos][test[ix][removePos]];
		}
		return getScore(Constants.TUNE_SET, scoreFunction, tuneEval);
	}// end removeAttribute(int, int)

	/**
	 * This method replaces an attribute already in the model. It relearns a NB
	 * model incorporating this feature in place of a previous feature, then
	 * evaluates the new network on the tune set according to the score function
	 * the user selected.
	 * 
	 * @param trainPosVector
	 *            value of new attribute for the pos train examples
	 * @param trainNegVector
	 *            value of new attribute for the neg train examples
	 * @param tunePosVector
	 *            value of new attribute for the pos tune examples
	 * @param tuneNegVector
	 *            value of new attribute for the neg tune examples
	 * @param testPosVector
	 *            value of new attribute for the pos test examples
	 * @param testNegVector
	 *            value of new attribute for the neg test examples
	 * @param replacePredNum
	 *            the predicate number to replaced in the model
	 * @param scoreFunction
	 *            metric for scoring the new network
	 * @param numValues
	 *            the number of values the new attribute can take
	 * @return the score of the model incorporating this feature
	 */
	public double replaceAttribute(byte[] trainPosVector,
			byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector,
			byte[] testPosVector, byte[] testNegVector, int replacePredNum,
			int scoreFunction, int numValues) {
		modifiedV1 = predNameMap.get(replacePredNum);
		int replacePos = modifiedV1;
		modPred1 = replacePos;

		replaceEvaluated++;
		moveType = Constants.REPLACE;
		numValuesFeat[numRules] = numValues;
		// get the conditional counts and treat as if adding new rule
		getCondCount(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector);

		ratio[numRules] = new double[numValues];
		for (int ix = 0; ix < numValues; ix++) {
			ratio[numRules][ix] = logCondProbs[numRules][ix][Constants.NEG]
					- logCondProbs[numRules][ix][Constants.POS];
		}
		for (int ix = 0; ix < numValues; ix++) {
			ratio[numRules][ix] = logCondProbs[numRules][ix][Constants.NEG]
					- logCondProbs[numRules][ix][Constants.POS];
		}
		for (int ix = 0; ix < train.length; ix++) {
			trainEval[ix] = trainLikelihood[ix]
					- ratio[replacePos][train[ix][replacePos]]
					+ ratio[numRules][train[ix][numRules]];
		}
		for (int ix = 0; ix < tune.length; ix++) {
			tuneEval[ix] = tuneLikelihood[ix]
					- ratio[replacePos][tune[ix][replacePos]]
					+ ratio[numRules][tune[ix][numRules]];
		}
		for (int ix = 0; ix < test.length; ix++) {
			testEval[ix] = testLikelihood[ix]
					- ratio[replacePos][test[ix][replacePos]]
					+ ratio[numRules][test[ix][numRules]];
		}

		return getScore(Constants.TUNE_SET, scoreFunction, tuneEval);
	}// end replaceAttribute(byte[], byte[], byte[], byte[], byte[], byte[],
		// int, int, int)

	/**
	 * Creates the condition distribution of this attribute based on the class
	 * variable (note all the training data is used to compute the
	 * distribution).
	 * 
	 * @param trainPosVector
	 *            the bitmap for the positive training examples
	 * @param trainNegVector
	 *            the bitmap for the negative training examples
	 * @param tunePosVector
	 *            the bitmap for the positive tune examples
	 * @param tuneNegVector
	 *            the bitmap for the negative tune examples
	 * @param testPosVector
	 *            the bitmap for the positive test examples
	 * @param testNegVector
	 *            the bitmap for the negative test examples
	 */
	protected void getCondCount(byte[] trainPosVector, byte[] trainNegVector,
			byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector,
			byte[] testNegVector) {
		// add attritube to the train, tune and test sets. compute counts
		super.getCondCount(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector);

		// compute the log p
		logCondProbs[numRules] = new double[numValuesFeat[numRules]][2];
		for (int ix = 0; ix < condCnts[0].length; ix++) {
			double total = 0;
			// get totals
			for (int jx = 0; jx < condCnts.length; jx++) {
				total += condCnts[jx][ix];
			}
			// calculate the log conditional probabilities
			for (int jx = 0; jx < condCnts.length; jx++) {
				logCondProbs[numRules][jx][ix] = Math.log(condCnts[jx][ix]
						/ total);
			}
		}// end of ix loop for getting totals/probs
	}// end of method getCond

	/**
	 * This method permanently adds the most recently introduced feature into
	 * feature set. Updates the cpts and network topology to reflect the change.
	 */
	public void commit() {
		long curTime = System.currentTimeMillis();
		long tmp = (curTime - startTime);
		long minutes = tmp / (long) 60000;
		double secs = Utils.round(((tmp - 60000 * minutes) / 1000.0), 5);
		if (moveType == Constants.ADD) {
			System.out.print("Added rule: ");
			// increment the number of rules in the network
			numRules++;
			numAddAccepted++;
		} else if (moveType == Constants.REMOVE) {
			System.out.print("Removed rule: ");
			// decrement the number of rules in the network
			numRules--;
			invPredName.remove(numRules);
			predNameMap.remove(modPred1);
			numRemoveAccepted++;
			if (modifiedV1 != numRules) {
				invPredName.put(modifiedV1, predNumber[numRules]);
				predNameMap.put(predNumber[numRules], modifiedV1);
			}
		} else if (moveType == Constants.REPLACE) {
			System.out.print("Replaced rule: ");
			numReplaceAccepted++;
		} else if (moveType == Constants.DISJOIN) {
			numRules--;
		}
		if (moveType == Constants.REPLACE
				|| (moveType == Constants.REMOVE && modifiedV1 != numRules)) {
			for (int ix = 0; ix < train.length; ix++) {
				train[ix][modifiedV1] = train[ix][numRules];
			}
			for (int ix = 0; ix < tune.length; ix++) {
				tune[ix][modifiedV1] = tune[ix][numRules];
			}
			for (int ix = 0; ix < test.length; ix++) {
				test[ix][modifiedV1] = test[ix][numRules];
			}
			ratio[modifiedV1] = ratio[numRules];
			logCondProbs[modifiedV1] = logCondProbs[numRules];
			predNumber[modifiedV1] = predNumber[numRules];
			headType[modifiedV1] = headType[numRules];

		}
		System.out.print(minutes + " mins " + secs + " secs ");
		System.out.println("rules in network " + numRules + " added by SAYU "
				+ (numRules - numPriorFeats) + " evaluated "
				+ (addEvaluated + removeEvaluated + replaceEvaluated));

		System.arraycopy(trainEval, 0, trainLikelihood, 0, train.length);
		System.arraycopy(tuneEval, 0, tuneLikelihood, 0, tune.length);
		System.arraycopy(testEval, 0, testLikelihood, 0, test.length);

		moveType = -1;
		modifiedV1 = -1;
		modifiedV2 = -1;
		modPred1 = -1;
		modPred2 = -1;
		if (numRules == (initSize - 1)) {
			doubleSizes();
		}

	}// end commit()

	/**
	 * Scores the current model on either train, tune or test set. Currently,
	 * the following score functions are available: CLL, AUC-PR, AUC-ROC and
	 * Accuracy.
	 * 
	 * @param dataSet
	 *            whether to score the train, tune or test set
	 * @param scoreFunction
	 *            the metric to use for scoring the classifier
	 * @return score of the model
	 */
	public double getScore(int dataSet, int scoreFunction) {
		if (dataSet == Constants.TRAIN_SET) {
			return getScore(dataSet, scoreFunction, trainLikelihood);
		} else if (dataSet == Constants.TUNE_SET) {
			return getScore(dataSet, scoreFunction, tuneLikelihood);
		} else if (dataSet == Constants.TEST_SET) {
			return getScore(dataSet, scoreFunction, testLikelihood);
		}
		return -1;
	}// end of method getScore(int, int)

	/**
	 * Score a model on either train, tune or test set. The model could be a
	 * speculative model. Currently, the following score functions are
	 * available: CLL, AUC-PR, AUC-ROC and Accuracy.
	 * 
	 * @param dataSet
	 *            whether to score the train, tune or test set.
	 * @param scoreFunction
	 *            the metric to use for scoring the classifier.
	 * @param scores
	 * @return the current score of the model
	 */
	protected double getScore(int dataSet, int scoreFunction, double[] scores) {
		double[] probs = null;
		int posCnt = trainPos;
		int negCnt = trainNeg;
		int idx = 0;
		if (dataSet == Constants.TRAIN_SET) {
			probs = getProbs(scores);
		} else if (dataSet == Constants.TUNE_SET) {
			probs = getProbs(scores);
			posCnt = tunePos;
			negCnt = tuneNeg;
			idx = 1;
		} else if (dataSet == Constants.TEST_SET) {
			probs = getProbs(scores);
			posCnt = testPos;
			negCnt = testNeg;
			idx = 2;
		}
		if (scoreFunction == Constants.ACC) {
			// FLAG
			int[] roc = ScoreFunctions.fastAccuracy(probs, cats[idx], 0.5);
			return ((double) (roc[Constants.TP] + roc[Constants.TN]))
					/ ((double) (probs.length));
		} else if (scoreFunction == Constants.PR_AREA) {
			double[] areas = ScoreFunctions.calcPRArea(probs, cats[idx],
					posCnt, negCnt);
			return areas[1];
		} else if (scoreFunction == Constants.ROC_AREA) {
			return ScoreFunctions.calcROCArea(probs, cats[idx]);
		} else if (scoreFunction == Constants.CLL) {
			return ScoreFunctions.cll(probs, cats[idx]);
		} else if (scoreFunction == Constants.X_ENT) {
			return -1;// ScoreFunctions.crossEntropyError(probs, cats[idx]);
		} else if (scoreFunction == Constants.MSQE) {
			return -1;// ScoreFunctions.msqe(probs, cats[idx]);
		} else if (scoreFunction == Constants.LIFT_AREA) {
			return Uplift.computeLiftAUC(probs, this.cats[idx]);
		}
		return -1;
	}// end getScore(int, int, ProbDistribution[], int)

	/**
	 * Calculates the probability that each example in either the train, tune or
	 * test set belongs to the positive class. It uses the current model
	 * structure.
	 * 
	 * @param dataSet
	 *            whether to return the probabilities for the train, tune or
	 *            test set.
	 * @return the Prob(Pos Class) of each example in the desired dataSet
	 */
	public double[] getProbs(int dataSet) {

		if (dataSet == Constants.TRAIN_SET) {
			return getProbs(trainLikelihood);
		} else if (dataSet == Constants.TUNE_SET) {
			return getProbs(tuneLikelihood);
		} else if (dataSet == Constants.TEST_SET) {
			return getProbs(testLikelihood);
		}
		System.out.println("Data set is invalid");
		return null;
	}// end getProbs(int)

	/**
	 * @param scores
	 *            the log odds (prob(neg)/prob(pos)) of each example.
	 * @return the probability that each example in scores is positive
	 */
	protected double[] getProbs(double[] scores) {
		double prob;
		double[] probabilities = new double[scores.length];
		for (int ix = 0; ix < scores.length; ix++) {
			prob = Math.exp(scores[ix]);
			prob = 1 / (1 + prob);
			probabilities[ix] = prob;
		}
		return probabilities;
	}// end getProbs(double[])

	/**
	 * Doubles the sizes of all of the instance variables.
	 */
	protected void doubleSizes() {
		super.doubleSizes();
		double[][][] newLogCondProbs = new double[initSize][][];
		double[][] newRatio = new double[initSize][];
		for (int ix = 0; ix < numRules; ix++) {
			newLogCondProbs[ix] = logCondProbs[ix];
			newRatio[ix] = ratio[ix];
		}
		logCondProbs = newLogCondProbs;
		ratio = newRatio;
	}
}