package sayu;

import java.io.*;
import java.util.*;
import java.util.regex.*;

/**
 * Java class to compute the Area Under the Uplift Curve
 * 
 * @author: Houssam Nassif
 * @version: 2
 */
public class Uplift {
	// constants. Maybe change to boolean?
	public static final int POS_CLASS = 1; // do not change, see
											// computeUpliftAUC
	public static final int NEG_CLASS = 0;
	public static final int TARGET_STRAT = 1;
	public static final int OTHER_STRAT = 0;

	// array to hold the instance values for curve construction
	List<LiftNode> targetList, otherList;

	/**
	 * Class constructor. Initializes the internal data structure.
	 * 
	 * @param score
	 *            a {@link Double} array of scores
	 * @param label
	 *            an {@link Integer} array of class labels, 1 for positives and
	 *            0 for negatives
	 * @param stratum
	 *            an {@link Integer} array of stratum markers, 1 for target
	 *            stratum and 0 for other stratum
	 */
	public Uplift(Double[] score, Integer[] label, Integer[] stratum) {
		targetList = new ArrayList<LiftNode>();
		otherList = new ArrayList<LiftNode>();

		for (int i = 0; i < score.length; i++) {
			if (stratum[i] == TARGET_STRAT)
				targetList.add(new LiftNode(score[i], label[i]));
			else
				otherList.add(new LiftNode(score[i], label[i]));
		}
	}

	/**
	 * Computes the lift AUC for both target and other strata Returns the uplift
	 * AUC = liftAUC(target) - liftAUC(other)
	 * 
	 * @return double uplift AUC score
	 */
	public double computeUpliftAUC() {
		return Uplift.computeLiftAUC(targetList) -
				Uplift.computeLiftAUC(otherList);
	}
	
	/**
	 * Computes the lift AUC of a given set of probabilities and labels,
	 * assuming that they are all for the same stratum.
	 * @param probs the probabilities given to the examples
	 * @param labels the class labels, 1 for positive 0 for negative
	 * @return
	 */
	public static double computeLiftAUC(double[] probs, int[] labels) {
		List<LiftNode> nodes = new ArrayList<LiftNode>();
		for (int i = 0; i < probs.length; i++) {
			nodes.add(new LiftNode(probs[i], labels[i]));
		}
		return Uplift.computeLiftAUC(nodes);
	}

	/**
	 * Computes the lift AUC of a given list Sorts the lists in descending order
	 * Uses trapezoidal rule to compute AUC Returns the area scaled to a
	 * fractional X-axis [0,1].
	 * 
	 * @param list
	 *            The List<LiftNode> data
	 * @return double lift AUC score
	 */
	public static double computeLiftAUC(List<LiftNode> list) {
		// sort data by their descending score
		Collections.sort(list, Collections.reverseOrder());

		// initialize variables
		int nbPos, nbInstance, previousNbPos, previousNbInstance, sum;
		nbPos = nbInstance = previousNbPos = previousNbInstance = sum = 0;

		//System.out.println("[" + list.get(0).getScore() + "," + list.get(list.size() - 1).getScore() + "]");
		// incrementaly compute lift AUC
		for (int i = 0; i < list.size(); i++) {
			nbInstance++; // one more instance
			nbPos += list.get(i).getLabel(); // assuming POS_CLASS is not
												// changed

			// look ahead to see if end of stretch of same score
			if (i < list.size() - 1
					&& Math.abs(list.get(i).getScore() - list.get(i + 1).getScore()) < 0.000001)
				continue;

			// not same stretch, compute lift AUC for this point increment
			// use trapezoidal rule
			sum += (nbInstance - previousNbInstance) * (nbPos + previousNbPos);
			//if (list.size() < 60) {
			//	System.out.println(nbInstance + " " + previousNbInstance + " " + nbPos + " " + previousNbPos);
			//}
			previousNbInstance = nbInstance;
			previousNbPos = nbPos;

		}
		//System.exit(0);

		// divide by nbInstance so that x axis is [0,1]
		return sum / (2.0 * nbInstance);
	}

	// Nodes to store score and class of instance
	static class LiftNode implements Comparable<LiftNode> {
		private Double score;
		private Integer label;

		public LiftNode(Double score, Integer label) {
			this.score = score;
			this.label = label;
		}

		public Double getScore() {
			return score;
		}

		public Integer getLabel() {
			return label;
		}

		// natural ordering
		public int compareTo(LiftNode obj) {
			Double score2 = obj.getScore();
			if (score > score2)
				return 1;
			else if (score < score2)
				return -1;
			else
				return 0;
		}
	} // end class LiftNode
	
	/**
	 * Main method, used to test the class. Will be ignored if class is
	 * instantiated. Takes two files, one for each stratum, and prints the
	 * uplift AUC score Usage example: java Uplift DP_OS.prob DP_YS.prob
	 * 
	 * @param File
	 *            name of target class, in Aleph .f format (ex DP_OS.prob)
	 * @param File
	 *            name of other class, in Aleph .f format (ex DP_YS.prob)
	 * @return Nothing, prints the uplift AUC score
	 */
	public static void main(String args[]) throws java.io.IOException {
		// Lists holding score, class label and stratum of each instance
		List<Double> score = new ArrayList<Double>();
		List<Integer> label = new ArrayList<Integer>();
		List<Integer> stratum = new ArrayList<Integer>();

		// Read and process both input files
		for (int i = 0; i <= 1; i++) {
			// Read file line by line
			BufferedReader br = new BufferedReader(new FileReader(args[i]));
			String strLine;

			// get the class and score of each instance
			Pattern p = Pattern
					.compile(".*(neg|pos),invasive\\(\\d+\\),(\\d+\\.\\d+)");
			while ((strLine = br.readLine()) != null) {
				Matcher m = p.matcher(strLine);
				if (m.find()) {
					// add the instance value to arrays
					score.add(Double.parseDouble(m.group(2)));
					if (m.group(1).equals("pos"))
						label.add(POS_CLASS);
					else
						label.add(NEG_CLASS);
					if (i == 0)
						stratum.add(TARGET_STRAT);
					else
						stratum.add(OTHER_STRAT);
				}
			}
		}

		// call constructor
		Uplift uplift = new Uplift(score.toArray(new Double[0]),
				label.toArray(new Integer[0]), stratum.toArray(new Integer[0]));
		System.out.println("Uplift is: " + uplift.computeUpliftAUC());
	}

} // end class Uplift
