>%
% This module implements a simple bottom clause generator.
%
%%
% It relies on 3 data structures:
% - a try with all goals launched so far
% - a domain(pred,value) dynamic predicate that stores all known values in the domain
% - a bc(Id,Level,Atom) with ground versions of the atoms. 
%
%
% Interface:
%  sat/1: saturate positive example Id.
%
%  sat/2: saturate example Id of type Type.
%
%  bottom_clause_to_clause/1: generate bottom-clause as clause.
%
%  next_literal/3: given a sequence of literals S, 
%      generate S' and respective clause Clause (variabilised).
%
%  get_clause_beam/1: final class to retrieve a node from Beam.
%
% Global Variables:
%
% i: depth
%
%


:- module(vista_bottom_clause,[sat/1,
	sat/2,
	bottom_clause_to_clause/1,
	next_literal/3,
	constant_histogram/0,
	hist/4,
	constant_from_bcs/3,
	ensure_sat/2,
	get_bc_literal/2,
	constants_from_an_example/4,
	has_hist/0]).

:- use_module(modes,[has_modeb/2,
	has_modeh/2,
	modeb/1,
	head_type/2,
	goal_mode/2,
	determining/2]).

:- use_module(examples,[example/3]).

:- use_module(hierarchy,[ok_type/1]).

:- use_module(library(tries),[trie_check_entry/3,
	trie_put_entry/3,
	trie_open/1,
	trie_close/1]).

:- use_module(library(ordsets),[ord_union/3,
	ord_intersection/3]).

:- use_module(library(rbtrees),[rb_new/1,
	rb_lookup/3,
	rb_insert/4]).

:- use_module(library(lists),[nth/3,member/2,sumlist/2]).

:- use_module('../common/utils',[list_to_conjunction/2]).

:- dynamic i/1,
	domain/2,
	bc/3,
	hist/4,
	has_hist/0,
	found/2.

:- use_module(library(dbusage)).

i(3).

% a cache to quickly find constants for the last bottom clause.
:- dynamic constants/2, current_sat/2.

sat(ExId) :-
	sat(ExId, pos).


sat(ExId, PosNeg) :-
	init(ExId, PosNeg, Goals),
%	db_dynamic,
	i(Level),
	loop(1, Level, Goals),
	trie_close(Goals).

loop(Level, Level, _Goals) :-!.
loop(Level, _, Goals) :-
	next_layer_bc(Goals, Level),
	fail.
loop(Level0, LevelF, Goals) :-
	Level1 is Level0+1,
	loop(Level1, LevelF, Goals).

next_layer_bc(InpGoals, Level) :-
	setof(Goal, InpGoals^add_goal(InpGoals, Goal), Goals),
	copy_to_domain(Goals),
	add_literals_to_bcl(Goals, Level).

add_goal(InpGoals, Goal) :-
	new_goal(Goal, Mode, InpGoals),
	answers(Goal, Mode).

%
% enumerate all goals from previous layers.
%
new_goal(Goal, Mode, InpGoals) :-
	modeb(Mode),
	predicate_property(user:Mode,number_of_clauses(N)),
	N > 0,
	fill_input_args(Mode, Goal),
	\+ trie_check_entry(InpGoals,Goal,_),
	trie_put_entry(InpGoals,Goal,_).

fill_input_args(Mode, Goal) :-
	Mode =.. [Name|Modes],
	new_list(Modes, Args),
	Goal =.. [Name|Args],
	fill_args(Modes, Args).

fill_args([], []).
fill_args(Mode.Modes, Arg.Args) :-
	fill_arg(Mode, Arg),
	fill_args(Modes, Args).

%
% use current domain...
%
fill_arg(+Type, Arg) :-
	domain(Type, Arg).
fill_arg(-_, _).
fill_arg(#_, _).

new_list([],[]).
new_list(_.Modes, _.Args) :-
	new_list(Modes, Args).

answers(Goal, Mode) :-
	has_modeb(Determinism, Mode),
	call_goal(Determinism, Goal).

call_goal(*, Goal) :-
	call(user:Goal).
call_goal(1, Goal) :-
	once(user:Goal).
call_goal(n, _Goal) :-
	throw(unsupported_yet(determinism=n)).

init(ExId, PosNeg, Goals) :-
	init_globals(Goals),
	example(ExId, PosNeg, Ex),
	copy_to_domain_types([Ex]),
	add_to_bcl(Ex, 0).

init_globals(Goals) :-
	trie_open(Goals),
	retractall(domain(_,_)),
	retractall(bc(_,_,_)),
	retractall(found(_,_)),
	nb_setval(bc_counter, 0).
	
copy_to_domain([]).
copy_to_domain([Goal|_Goals]) :-
	goal_mode(Goal, Modes),
	Goal =.. [_|Args],
	store_to_domains(Modes, Args),
	fail.
copy_to_domain([_|Goals]) :-
	copy_to_domain(Goals).

store_to_domains([], []).
store_to_domains(+_.Types, _.Goals) :-
	store_to_domains(Types, Goals).
store_to_domains(#_.Types, _.Goals) :-
	store_to_domains(Types, Goals).
store_to_domains(-Type.Types, Arg.Goals) :-
	store_to_domain(Type, Arg),
	store_to_domains(Types, Goals).

%
% same as before, but used for modeh, and needs to copy everything.
%
copy_to_domain_types([]).
copy_to_domain_types([Goal|_Goals]) :-
	head_type(Goal, Modes),
	Goal =.. [_|Args],
	store_to_domain_types(Modes, Args),
	fail.
copy_to_domain_types([_|Goals]) :-
	copy_to_domain_types(Goals).

store_to_domain_types([], []).
store_to_domain_types(Type.Types, Goal.Goals) :-
	store_to_domain(Type, Goal),
	store_to_domain_types(Types, Goals).

store_to_domain(Type, Goal) :-
	domain(Type, Goal), !.
store_to_domain(Type, Goal) :-
	assert(domain(Type, Goal)).

add_literals_to_bcl(Goal.Goals, Level) :-
	add_to_bcl(Goal, Level),
	add_literals_to_bcl(Goals, Level).
add_literals_to_bcl([], _).

add_to_bcl(Ex, _Level) :-
	bc(_, _, Ex), !.
add_to_bcl(Ex, Level) :-
	nb_getval(bc_counter, I0),
	I is I0+1,
	nb_setval(bc_counter, I),
	assert(bc(I, Level, Ex)).

get_bc_literal(Id, G) :-
	bc(Id, _Level, G).

%
% construct a clause from the current bottom database.
%
bottom_clause_to_clause((Head :- Body)) :-
	rb_new(Map0),
	make_head(Head, MapH, Map0),
	make_body(Body, _, MapH).

make_head(Head, MapH, Map0) :-
	bc(1,0,Ground),
	head_type(Head, Types),
	Ground =.. [Name|Args],
	to_map_from_types(Args, Vars, Types, MapH, Map0),
	Head =.. [Name|Vars].

to_map_from_types([], [], [], Map, Map).
to_map_from_types(Arg.Args, Var.Vars, Type.Types, MapH, Map) :-
	domain(Type, Arg),
	to_map_from_type(Arg, Var, Type, MapHI, Map),
	to_map_from_types(Args, Vars, Types, MapH, MapHI).

to_map_from_type(Arg, Var, Type, Map, Map) :-
	rb_lookup(Arg:Type, Var, Map), !.
to_map_from_type(Arg, Var, Type, MapHi, Map) :-
	rb_insert(Map, Arg:Type, Var, MapHi).

make_body(Body, MapB, MapH) :-
	nb_getval(bc_counter, Max),
	Max > 0, !,
	make_body(1, Max, MapB, MapH, BodyList, []),
	list_to_conjunction(BodyList, Body).
make_body(true, Map, Map).

make_body(I, I, Map, Map) --> !.
make_body(I, Max, MapB, MapB0) -->
	{
	  I1 is I+1,
	  bc(I1,_,Ground),
	  findall(Mode, goal_mode(Ground, Mode), Modes)
	},
	make_literal(Modes, Ground, MapB1, MapB0),
	make_body(I1, Max, MapB, MapB1).

%
% a literal may have several modes
%
make_literal([], _, MapB, MapB) --> [].
make_literal(Mode.Modes, Ground, MapB1, MapB0) -->
	{ make_body_goal(Mode, Ground, MapB2, MapB0, Literal) }, !,
	[ Literal ],
	make_literal(Modes, Ground, MapB1, MapB2).
make_literal(_.Modes, Ground, MapB1, MapB0) -->
	make_literal(Modes, Ground, MapB1, MapB0).

make_body_goal(LMode, Ground, MapB2, MapB0, Literal) :-
	Ground =.. [P|Args],
	make_literal(LMode, Args, NewArgs, MapB2, MapB0),
	Literal =.. [P|NewArgs].

make_literal([], [], [], MapB, MapB).
make_literal(Mode.MArgs, Arg.Args, New.NewArgs, MapBF, MapB0) :-
	make_arg(Mode, Arg, New, MapBI, MapB0), !,
	make_literal(MArgs, Args, NewArgs, MapBF, MapBI).

%
% be careful to make sure Arg is of correct type 
%
make_arg(+Type, Arg, V, MapBI, MapB0) :-
	domain(Type, Arg),
	to_map_from_type(Arg, V, Type, MapBI, MapB0).
make_arg(-Type, Arg, V, MapBI, MapB0) :-
	domain(Type, Arg),
	to_map_from_type(Arg, V, Type, MapBI, MapB0).
make_arg(#_, Arg, Arg, MapB, MapB).



%
% Given a list of literals of the form id-modeargs, get the next matching literal
% from the data-base, in the form Id and mode.
%
% We also get the variabilised clause, kind of as side-effect.
%
next_literal(Literals, NewLiterals, (Head :- Clause)) :-
	rb_new(Map0),
	make_head(Head, MapH, Map0),
	add_literals(Literals, NewLiterals, [Id-Mode], Position, 1, Map, MapH, ClauseLiterals, Last),
	expand_literal(Position, Id-Mode, Map, Last),
	list_to_conjunction(ClauseLiterals, Clause).

add_literals([], Ls, Ls, P, P, Map, Map) --> [].
add_literals(Literal.Literals, Literal.New, MoreNew, PF, P0, Map, Map0) -->
	add_literal(Literal, PI, P0, MapI, Map0),
	add_literals(Literals, New, MoreNew, PF, PI, Map, MapI).

add_literal(I-LMode, PI, P0, MapI, Map0) -->
	{ bc(I, _, Ground),
	  make_body_goal(LMode, Ground, MapI, Map0, Literal),
          ( I > P0 -> PI = I ; PI = P0 )
        },
	[Literal].

expand_literal(Position, I-LMode, Map, [Literal]) :-
	nb_getval(bc_counter, Max),
	P is Position+1,
	between(P, Max, I),
	bc(I, _, Ground),
	check_input(Ground, LMode, Map),
	make_body_goal(LMode, Ground, _, Map, Literal).

check_input(Ground, LMode, Map) :-
	goal_mode(Ground, LMode),
	fulfill_input(LMode, 1, Ground, Map).

fulfill_input([], _, _, _).
fulfill_input([+Type|LMode], I, Ground, Map) :-
	arg(I, Ground, Arg),
	rb_lookup(Arg:Type, _, Map),
	I1 is I+1,
	fulfill_input(LMode, I1, Ground, Map).
fulfill_input([-_|LMode], I, Ground, Map) :-
	I1 is I+1,
	fulfill_input(LMode, I1, Ground, Map).
fulfill_input([#_|LMode], I, Ground, Map) :-
	I1 is I+1,
	fulfill_input(LMode, I1, Ground, Map).

constant_histogram :-
	retract(has_hist),
	fail.
constant_histogram :-
	retractall(hist(_,_,_,_)),
	fail.
constant_histogram :-
	PosNeg = pos,
	example(Id, PosNeg, Ex),
	\+ ( example(Id1, PosNeg, Ex), Id1 < Id),
	sat(Id, PosNeg),
	add_constants(PosNeg, Id),
	fail.
constant_histogram :-
	assert(has_hist),
	fail.
constant_histogram.  :-
	listing(hist).

pos_constant_histogram :-
	retractall(has_hist),
	fail.
pos_constant_histogram :-
	retractall(hist(_,_,_,_)),
	fail.
pos_constant_histogram :-
	example(Id, pos, Ex),
	\+ ( example(Id1, neg, Ex), Id1 < Id),
	sat(Id),
	add_constants(pos, Id),
	fail.
pos_constant_histogram :-
	assert(has_hist),
	fail.
pos_constant_histogram(_). % :-
%	listing(hist).

add_constants(Polarity, _Id) :-
	determining(_,Pred),
	goal_mode(Pred, Modes),
	nth(N, Modes, #Type),
	bc(_,_,Pred),
	arg(N, Pred, C),
	add_c(Type, C, Polarity),
	fail.
add_constants(_Polarity, _Id).

add_c(Type, C, Polarity) :-
	retract(hist(Type, C, Polarity, N)), !,
	N1 is N+1,
	assert(hist(Type, C, Polarity, N1)).
add_c(Type, C, Polarity) :-
	assert(hist(Type, C, Polarity, 1)).

constant_from_bcs(Type, N, C) :-
%	listing(has_hist),
	has_hist, !,
	constant_in_histogram(Type, N, C).
constant_from_bcs(Type, N, C) :-
	constant_histogram,
	constant_from_bcs(Type, N, C).

constant_in_histogram(Type, N, C) :-
	hist(Type, C, pos, N1), N1 >= N.

constants_from_an_example(Id, PosNeg, Type, Constants) :-
	ensure_sat(Id, PosNeg),
	cached_constants(Type, Constants).

ensure_sat(Id, Pos) :-
	current_sat(Id, Pos), !.
ensure_sat(Id, Pos) :-
	retractall(constants(_,_)),
	retractall(current_sat(_,_)),
	sat(Id, Pos),
	assert(current_sat(Id,Pos)).

cached_constants(Type, Constants) :-
	constants(Type, Constants), !.
cached_constants(Type, Constants) :-
	setof(Const, find_constant(Type, Const), Constants),
	assert(constants(Type, Constants)).

find_constant(Type, Const) :-
	determining(_,Pred),
	once(
	  (goal_mode(Pred, Modes),
	   nth(_, Modes, #Type)
          )
        ),
	nth(N, Modes, #Type),
	bc(_,_,Pred),
	arg(N, Pred, Const).

