
:- module(vista_foil,[next_clause/1, init_foil/1]).

:- use_module('../common/utils',[clause_length/3,
		     remove_dups/2,
		     randomize_list/2]).

:- use_module('../common/debug_write',[debug_format/3]).

:- use_module('mdie',[generate_mdie_clause/2]).

:- use_module(variabilise,[variables/2]).

:- use_module(literals,[all_literals/2]).

:- use_module(constants,[constants_from_bottom_clauses/1]).

:- use_module(modes,[modeb/1]).

:- use_module(search_space,[get_node_from_search_space/1]).

:- use_module(library(lists),[member/2]).

:- dynamic use_negation/1.

use_negation(off).

init_foil(Head) :-
	recorda(clause, (Head), _).

%
% Clauses from the last refinement step
%
next_clause(Clause) :-
	recorded(clause,Clause,DBRef), !,
	erase(DBRef).
%
% Unexpanded nodes
%
next_clause(NewClause) :-
	%jjd beam search
	get_node_from_search_space(Clause),
	clause_length(Clause,0,ClLength),
	( vista_search_space:clauselength(Max), Max > ClLength+1 ->
	    !
	;
	    fail
	),
%	(Clause = (_ :- _) -> trace ; true),
	generate_clause(Clause, NewClause).


generate_clause(Clause, NewClause) :-
	constants_from_bottom_clauses(true), !,
	generate_all_mdie_clauses(Clause, NewClause).
generate_clause(Clause, NewClause) :-
        variables(Clause, Variables),
        all_literals(Variables, Literals),
        remove_dups(Literals, RemainingLiterals),
	randomize_list(RemainingLiterals, RandomLiterals),
	debug_format(2,'refine ~w -> ~w~n',[Clause,RandomLiterals]),
	(
	  generate_all_refinements(Clause, RandomLiterals)
	;
	  next_clause(NewClause)
	).

generate_all_mdie_clauses(Clause, NewClause) :-
	findall(NewClause,do_generate_mdie_clause(Clause, NewClause),NewClauses0),
%	randomize_list(NewClauses0, NewClauses),
	NewClauses0 = NewClauses,
	member(NewClause, NewClauses),
	recordz(clause, NewClause, _),
	fail.
generate_all_mdie_clauses(_Clause, NewClause) :-
	next_clause(NewClause).

do_generate_mdie_clause(Clause, NewClause) :-
	generate_mdie_clause(Clause, NewClause),
	get_value(refinements,R0),
	R is R0+1,
	set_value(refinements,R).
	
generate_all_refinements(Clause, RandomLiterals) :-
	generate_refinements(Clause, RandomLiterals),
	fail.

% The following 3 predicates are for generating refinements
% of a clause. The refinements are stored in a recorded.
% The first argument is the clause, the second is a list of 
% instantiated literals that can be used to refine the clause.
%
%generate_refinements(_,[]) :-
%    !.

%
% Clause had both a head and a body
%
generate_refinements((Head :- Body), [Lit|_]) :-
    get_value(refinements,R0),
    R is R0+1,
    set_value(refinements,R),
    add_lit(Body, Lit, NBody),
    debug_format(2,'add ~w~n',[(Head :- NBody)]),
    recordz(clause,(Head :- NBody),_).
generate_refinements((Head :- Body), [Lit|_]) :-
    get_value(refinements,R0),
    R is R0+1,
    set_value(refinements,R),
    use_negation(on),
    can_negate(Lit, (Head :- Body)),
    add_lit(Body, \+ Lit, NBody),
    debug_format(2,'add ~w~n',[(Head :- NBody)]),
    recordz(clause,(Head :- NBody),_).
generate_refinements((Head :- Body), [_|Literals]) :- !,
    generate_refinements((Head :- Body), Literals).
%
% Clause only consists of a clause head
% 
generate_refinements(Head, [Lit|_]) :-
    get_value(refinements,R0),
    R is R0+1,
    set_value(refinements,R),
    debug_format(2,'add ~w~n',[(Head :- Lit)]),
    recordz(clause,(Head :- Lit),_).
generate_refinements(Head, [Lit|_]) :-
    get_value(refinements,R0),
    R is R0+1,
    set_value(refinements,R),
    use_negation(on),
    can_negate(Lit, (Head :- true)),
    debug_format(2,'add ~w~n',[(Head :- \+Lit)]),
    recordz(clause,(Head :- \+ Lit),_).
generate_refinements(Head, [_|Literals]) :-
    generate_refinements(Head, Literals).


add_lit((G,Body), Lit, (G,NBody)) :- !,
	add_lit(Body, Lit, NBody).
add_lit(Body, Lit, (Body,Lit) ).

% we can only perform negation if the modes of the literal
% are all input
can_negate(Lit,_) :-
    functor(Lit, Name, Arg),
    functor(ModeB, Name, Arg),
    modeb(ModeB),
    debug_format(2,'doing ~w/~w~n',[Lit,ModeB]),
    ModeB =.. [_|Modes],
    all_inp(Modes).

all_inp([]).
all_inp([+_|Modes]) :-
    all_inp(Modes).
all_inp([#_|Modes]) :-
    all_inp(Modes).


